from typing import Sequence, cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import TealInputError
from pyteal.ir import TealOp, Op, TealSimpleBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class NaryExpr(Expr):
    """N-ary expression base class.

    This type of expression takes an arbitrary number of arguments.
    """

    def __init__(
        self, op: Op, inputType: TealType, outputType: TealType, args: Sequence[Expr]
    ):
        super().__init__()
        if len(args) == 0:
            raise TealInputError("NaryExpr requires at least one child")
        for arg in args:
            if not isinstance(arg, Expr):
                raise TealInputError(
                    "Argument is not a PyTeal expression: {}".format(arg)
                )
            require_type(arg, inputType)
        self.op = op
        self.outputType = outputType
        self.args = args

    def __teal__(self, options: "CompileOptions"):
        start = None
        end = None
        for i, arg in enumerate(self.args):
            argStart, argEnd = arg.__teal__(options)
            if i == 0:
                start = argStart
                end = argEnd
            else:
                cast(TealSimpleBlock, end).setNextBlock(argStart)
                opBlock = TealSimpleBlock([TealOp(self, self.op)])
                argEnd.setNextBlock(opBlock)
                end = opBlock

        return start, end

    def __str__(self):
        ret_str = "(" + str(self.op).title().replace("_", "")
        for a in self.args:
            ret_str += " " + a.__str__()
        ret_str += ")"
        return ret_str

    def type_of(self):
        return self.outputType

    def has_return(self):
        return False


NaryExpr.__module__ = "pyteal"


def Add(*args: Expr) -> Expr:
    """Numerical addition.

    Produces the sum of all the input arguments.

    All arguments must be PyTeal expressions that evaluate to uint64, and there must be at least one
    argument.

    Example:
        ``Add(Int(1), Int(2), Int(3))``
    """
    return NaryExpr(Op.add, TealType.uint64, TealType.uint64, args)


def Mul(*args: Expr) -> Expr:
    """Numerical multiplication.

    Produces the product of all the input arguments.

    All arguments must be PyTeal expressions that evaluate to uint64, and there must be at least one
    argument.

    Example:
        ``Mul(Int(2), Int(3), Int(4))``
    """
    return NaryExpr(Op.mul, TealType.uint64, TealType.uint64, args)


def And(*args: Expr) -> Expr:
    """Logical and expression.

    Produces 1 if all arguments are nonzero. Otherwise produces 0.

    All arguments must be PyTeal expressions that evaluate to uint64, and there must be at least one
    argument.

    Example:
        ``And(Txn.amount() == Int(500), Txn.fee() <= Int(10))``
    """
    return NaryExpr(Op.logic_and, TealType.uint64, TealType.uint64, args)


def Or(*args: Expr) -> Expr:
    """Logical or expression.

    Produces 1 if any argument is nonzero. Otherwise produces 0.

    All arguments must be PyTeal expressions that evaluate to uint64, and there must be at least one
    argument.
    """
    return NaryExpr(Op.logic_or, TealType.uint64, TealType.uint64, args)


def Concat(*args: Expr) -> Expr:
    """Concatenate byte strings.

    Produces a new byte string consisting of the contents of each of the passed in byte strings
    joined together.

    All arguments must be PyTeal expressions that evaluate to bytes, and there must be at least one
    argument.

    Example:
        ``Concat(Bytes("hello"), Bytes(" "), Bytes("world"))``
    """
    return NaryExpr(Op.concat, TealType.bytes, TealType.bytes, args)



import pyteal as pt

from pyteal.compiler.sort import sortBlocks


def test_sort_single():
    block = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.addIncoming()
    block.validateTree()

    expected = [block]
    actual = sortBlocks(block, block)

    assert actual == expected


def test_sort_sequence():
    block5 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 5)])
    block4 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 4)])
    block4.setNextBlock(block5)
    block3 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 3)])
    block3.setNextBlock(block4)
    block2 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 2)])
    block2.setNextBlock(block3)
    block1 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])
    block1.setNextBlock(block2)
    block1.addIncoming()
    block1.validateTree()

    expected = [block1, block2, block3, block4, block5]
    actual = sortBlocks(block1, block5)

    assert actual == expected


def test_sort_branch():
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)
    block.addIncoming()
    block.validateTree()

    expected = [block, blockTrue, blockFalse]
    actual = sortBlocks(block, blockFalse)

    assert actual == expected


def test_sort_multiple_branch():
    blockTrueTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true true"')])
    blockTrueFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true false"')])
    blockTrueBranch = pt.TealConditionalBlock([])
    blockTrueBranch.setTrueBlock(blockTrueTrue)
    blockTrueBranch.setFalseBlock(blockTrueFalse)
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockTrueBranch)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)
    block.addIncoming()
    block.validateTree()

    expected = [
        block,
        blockTrue,
        blockTrueBranch,
        blockTrueFalse,
        blockTrueTrue,
        blockFalse,
    ]
    actual = sortBlocks(block, blockFalse)

    assert actual == expected


def test_sort_branch_converge():
    blockEnd = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.return_)])
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockEnd)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    blockFalse.setNextBlock(blockEnd)
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)
    block.addIncoming()
    block.validateTree()

    expected = [block, blockFalse, blockTrue, blockEnd]
    actual = sortBlocks(block, blockEnd)

    assert actual == expected



import pyteal as pt

options = pt.CompileOptions()


def assert_MaybeValue_equality(
    actual: pt.MaybeValue, expected: pt.MaybeValue, options: pt.CompileOptions
):
    actual_block, _ = actual.__teal__(options)
    actual_block.addIncoming()
    actual_block = pt.TealBlock.NormalizeBlocks(actual_block)

    expected_block, _ = expected.__teal__(options)
    expected_block.addIncoming()
    expected_block = pt.TealBlock.NormalizeBlocks(expected_block)

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual_block == expected_block

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual_block),
        pt.TealBlock.GetReferencedScratchSlots(expected_block),
    )


def test_maybe_value():
    ops = (
        pt.Op.app_global_get_ex,
        pt.Op.app_local_get_ex,
        pt.Op.asset_holding_get,
        pt.Op.asset_params_get,
    )
    types = (pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype)
    immedate_argv = ([], ["AssetFrozen"])
    argv = ([], [pt.Int(0)], [pt.Int(1), pt.Int(2)])

    for op in ops:
        for type in types:
            for iargs in immedate_argv:
                for args in argv:
                    expr = pt.MaybeValue(op, type, immediate_args=iargs, args=args)

                    assert expr.slotOk != expr.slotValue
                    assert expr.output_slots == [expr.slotValue, expr.slotOk]

                    assert expr.hasValue().type_of() == pt.TealType.uint64
                    with pt.TealComponent.Context.ignoreExprEquality():
                        assert expr.hasValue().__teal__(options) == pt.ScratchLoad(
                            expr.slotOk
                        ).__teal__(options)

                    assert expr.value().type_of() == type
                    with pt.TealComponent.Context.ignoreExprEquality():
                        assert expr.value().__teal__(options) == pt.ScratchLoad(
                            expr.slotValue
                        ).__teal__(options)

                    assert expr.type_of() == pt.TealType.none

                    expected_call = pt.TealSimpleBlock(
                        [
                            pt.TealOp(expr, op, *iargs),
                            pt.TealOp(None, pt.Op.store, expr.slotOk),
                            pt.TealOp(None, pt.Op.store, expr.slotValue),
                        ]
                    )

                    if len(args) == 0:
                        expected = expected_call
                    elif len(args) == 1:
                        expected, after_arg = args[0].__teal__(options)
                        after_arg.setNextBlock(expected_call)
                    elif len(args) == 2:
                        expected, after_arg_1 = args[0].__teal__(options)
                        arg_2, after_arg_2 = args[1].__teal__(options)
                        after_arg_1.setNextBlock(arg_2)
                        after_arg_2.setNextBlock(expected_call)

                    expected.addIncoming()
                    expected = pt.TealBlock.NormalizeBlocks(expected)

                    actual, _ = expr.__teal__(options)
                    actual.addIncoming()
                    actual = pt.TealBlock.NormalizeBlocks(actual)

                    with pt.TealComponent.Context.ignoreExprEquality():
                        assert actual == expected
